/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#if !defined(BOOST_TRAVERSAL_EXTEND_20070926_1425)
#define BOOST_TRAVERSAL_EXTEND_20070926_1425

#include <boost/traversal/functional/detail/callable_traits.hpp>
#include <boost/static_assert.hpp>

namespace boost { namespace traversal {
    namespace detail
    {
        template<typename Ret>
        struct extend_default
        {
            typedef Ret result_type;

            extend_default()
                : ret_()
            {}

            extend_default(Ret ret)
                : ret_(ret)
            {}

            template<typename U0>
            result_type
            operator()(U0&) const
            {
                return ret_;
            }

            template<typename U0, typename U1>
            result_type
            operator()(U0&, U1&) const
            {
                return ret_;
            }

            template<typename U0, typename U1, typename U2>
            result_type
            operator()(U0&, U1&, U2&) const
            {
                return ret_;
            }

            Ret ret_;
        };

        template<>
        struct extend_default<void>
        {
            typedef void result_type;

            template<typename U0>
            result_type
            operator()(U0&) const
            {
                return;
            }

            template<typename U0, typename U1>
            result_type
            operator()(U0&, U1&) const
            {
                return;
            }

            template<typename U0, typename U1, typename U2>
            result_type
            operator()(U0&, U1&, U2&) const
            {
                return;
            }
        };

        template<typename Func, int Arity>
        struct extend
        {
            BOOST_STATIC_ASSERT(Arity > 0 && Arity < 4);
        };

        template<typename Func>
        struct extend<Func, 1>
            : private extend_default<typename callable_result_type<Func>::type>
        {
            typedef extend_default<typename callable_result_type<Func>::type> base_type;
            using base_type::operator();

            typedef typename base_type::result_type result_type;
            typedef typename callable_argument_c<0, Func>::type argument_type;

            explicit extend(Func f)
                : f_(f)
            {}

            template<typename R>
            extend(Func f, R ret)
                : base_type(ret), f_(f)
            {}

            result_type
            operator()(argument_type a0) const
            {
                return f_(a0);
            }

            Func f_;
        };
        
        template<typename Func>
        struct extend<Func, 2>
            : private extend_default<typename callable_result_type<Func>::type>
        {
            typedef extend_default<typename callable_result_type<Func>::type> base_type;

            using base_type::operator();

            typedef typename base_type::result_type result_type;
            typedef typename callable_argument_c<0, Func>::type first_argument;
            typedef typename callable_argument_c<1, Func>::type second_argument;

            explicit extend(Func f)
                : f_(f)
            {}

            template<typename R>
            extend(Func f, R ret)
                : base_type(ret), f_(f)
            {}

            result_type
            operator()(first_argument a0, second_argument a1) const
            {
                return f_(a0, a1);
            }

            Func f_;
        };

        template<typename Func>
        struct extend<Func, 3>
            : private extend_default<typename callable_result_type<Func>::type>
        {
            typedef extend_default<typename callable_result_type<Func>::type> base_type;

            using base_type::operator();

            typedef typename base_type::result_type result_type;
            typedef typename callable_argument_c<0, Func>::type first_argument;
            typedef typename callable_argument_c<1, Func>::type second_argument;
            typedef typename callable_argument_c<2, Func>::type third_argument;

            explicit extend(Func f)
                : f_(f)
            {}

            template<typename R>
            extend(Func f, R ret)
                : base_type(ret), f_(f)
            {}

            result_type
            operator()(first_argument a0, second_argument a1, third_argument a2) const
            {
                return f_(a0, a1, a2);
            }

            Func f_;
        };
    }

    template<typename Func>
    detail::extend<Func, detail::callable_arity<Func>::value>
    extend(Func f)
    {
        return detail::extend<Func, detail::callable_arity<Func>::value>(f);
    }

    template<typename Func, typename Dflt>
    detail::extend<Func, detail::callable_arity<Func>::value>
    extend(Func f, Dflt dflt)
    {
        return detail::extend<Func, detail::callable_arity<Func>::value>(f, dflt);
    }
}}

#endif
