/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#if !defined(BOOST_TRAVERSAL_TU_EXTEND_20080315_1127)
#define BOOST_TRAVERSAL_EXTEND_20070926_1425

#include <boost/traversal/functional/detail/callable_traits.hpp>
#include <boost/static_assert.hpp>

namespace boost { namespace traversal {
    namespace detail
    {
        template<typename Func, int Arity>
        struct tu_extend
        {
            // Extend must handle both state and new value so arity > 1
            BOOST_STATIC_ASSERT(Arity > 1 && Arity < 4);
        };
        
        template<typename Func>
        struct tu_extend<Func, 2>
        {
            typedef typename callable_result_type<Func>::type result_type;
            typedef typename callable_argument_c<0, Func>::type first_argument;
            typedef typename callable_argument_c<1, Func>::type second_argument;

            explicit tu_extend(Func f)
                : f_(f)
            {}

            template<typename R>
            tu_extend(Func f, R ret)
                : f_(f)
            {}

            result_type
            operator()(first_argument a0, second_argument a1) const
            {
                return f_(a0, a1);
            }

            template<typename U0>
            result_type
            operator()(first_argument state, U0& u0) const
            {
                return state;
            }

            Func f_;
        };

        template<typename Func>
        struct tu_extend<Func, 3>
        {
            typedef typename callable_result_type<Func>::type result_type;

            typedef typename callable_argument_c<0, Func>::type first_argument;
            typedef typename callable_argument_c<1, Func>::type second_argument;
            typedef typename callable_argument_c<2, Func>::type third_argument;

            explicit tu_extend(Func f)
                : f_(f)
            {}

            template<typename R>
            tu_extend(Func f, R ret)
                : f_(f)
            {}

            result_type
            operator()(first_argument a0, second_argument a1, third_argument a2) const
            {
                return f_(a0, a1, a2);
            }

            template<typename U0, typename U1>
            result_type
            operator()(first_argument state, U0& u0, U1& u1) const
            {
                return state;
            }

            Func f_;
        };
    }

    template<typename Func>
    detail::tu_extend<Func, detail::callable_arity<Func>::value>
    tu_extend(Func f)
    {
        return detail::tu_extend<Func, detail::callable_arity<Func>::value>(f);
    }
}}

#endif
