/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/

#if !defined(BOOST_TRAVERSAL_FUSION_20080406_1704)
#define BOOST_TRAVERSAL_FUSION_20080406_1704

#include <boost/traversal/pattern/detail/all_predicates.hpp>

#include <boost/mpl/void.hpp>
#include <boost/mpl/and.hpp>
#include <boost/mpl/not.hpp>
#include <boost/mpl/int.hpp>
#include <boost/mpl/filter_view.hpp>
#include <boost/mpl/lambda.hpp>
#include <boost/mpl/equal_to.hpp>
#include <boost/mpl/size.hpp>
#include <boost/mpl/vector.hpp>

#include <boost/fusion/mpl.hpp>
#include <boost/fusion/support/is_sequence.hpp>
#include <boost/fusion/sequence/intrinsic/size.hpp>
#include <boost/fusion/adapted/mpl.hpp>

#include <boost/preprocessor/repetition/enum_params_with_a_default.hpp>
#include <boost/preprocessor/repetition/enum_params.hpp>

#include <boost/type_traits/is_same.hpp>

#if !defined(BOOST_TRAVERSAL_PATTERN_FUSION_MAX)
#define BOOST_TRAVERSAL_PATTERN_FUSION_MAX 10
#endif

namespace boost { namespace traversal { 

    namespace detail
    {
        template<typename PatternSeq, typename Seq>
        struct fusion_pattern_match
            : mpl::and_<
            boost::fusion::traits::is_sequence<Seq>,
            mpl::equal_to<mpl::size<PatternSeq>, mpl::size<Seq> >,
            detail::all_predicates<PatternSeq, Seq> >
        {};
    }

    namespace pattern {

    template<BOOST_PP_ENUM_PARAMS_WITH_A_DEFAULT(BOOST_TRAVERSAL_PATTERN_FUSION_MAX, typename T,mpl::void_)>
    struct fusion
    {
        template<typename Sequence>
        struct apply
            : detail::fusion_pattern_match<
            mpl::filter_view<
              mpl::vector<BOOST_PP_ENUM_PARAMS(10, T)>, 
              mpl::not_<is_same<mpl::void_, mpl::_> > >,
            Sequence>
        {};
    };
}}}

#endif
