/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/

#if !defined(BOOST_TRAVERSAL_VARIANT_20080406_1817)
#define BOOST_TRAVERSAL_VARIANT_20080406_1817

#include <boost/traversal/pattern/detail/all_predicates.hpp>
#include <boost/traversal/support/is_variant.hpp>

#include <boost/mpl/void.hpp>
#include <boost/mpl/and.hpp>
#include <boost/mpl/int.hpp>
#include <boost/mpl/not.hpp>
#include <boost/mpl/filter_view.hpp>
#include <boost/mpl/equal_to.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/mpl/list.hpp>
#include <boost/mpl/size.hpp>

#include <boost/variant.hpp>

#include <boost/preprocessor/repetition/enum_params_with_a_default.hpp>
#include <boost/preprocessor/repetition/enum_params.hpp>

#include <boost/type_traits/is_same.hpp>

namespace boost { namespace traversal { 

    namespace detail
    {
        template<typename Sequence, typename N>
        struct size_match
            : mpl::equal_to<
            mpl::size<typename Sequence::types>, N>
        {};

        template<typename Predicates, typename Variant>
        struct predicate_match
            : detail::all_predicates<Predicates, typename Variant::types>
        {};

        template<typename PatternSeq, typename Variant>
        struct variant_pattern_match
            : mpl::and_<
            traits::is_variant<Variant>,
            detail::size_match<Variant, mpl::size<PatternSeq> >,
            detail::predicate_match<PatternSeq, Variant> >
        {};
    }

    namespace pattern {

    template<BOOST_PP_ENUM_PARAMS_WITH_A_DEFAULT(BOOST_VARIANT_LIMIT_TYPES, typename T,mpl::void_)>
    struct variant
    {
        template<typename Variant>
        struct apply
            : detail::variant_pattern_match<
                mpl::filter_view<
                  mpl::vector<BOOST_PP_ENUM_PARAMS(10, T)>, 
                  mpl::not_<is_same<mpl::void_, mpl::_> > >,
                Variant>
        {};
    };
}}}

#endif
