/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/

#include <boost/traversal/recursive/full.hpp>
#include <boost/traversal/functional/extend.hpp>
#include <boost/fusion/adapted/struct/adapt_struct.hpp>

#include <vector>
#include <boost/variant.hpp>

using namespace boost;

namespace demo
{
    struct dept;

    enum currency
        {
            dollar,
            sterling,
            washers
        };

    struct salary
    {
        currency curr;
        int amount;
    };

    struct person
    {
        std::string name;
        std::string address;
    };

    struct employee
    {
        person detail;
        salary pay;
    };

    typedef variant<dept, employee> unit;

    struct dept
    {
        std::string name;
        std::vector<unit> units;
    };

    typedef std::vector<dept> company;

    bool increase_salary(salary& salary)
    {
        ++salary.amount;
        return false;
    }
}

BOOST_FUSION_ADAPT_STRUCT(demo::dept,
                          (std::string, name)
                          (std::vector<demo::unit>, units))

BOOST_FUSION_ADAPT_STRUCT(demo::employee,
                          (demo::person, detail)
                          (demo::salary, pay))

BOOST_FUSION_ADAPT_STRUCT(demo::person,
                          (std::string, name)
                          (std::string, address))

BOOST_FUSION_ADAPT_STRUCT(demo::salary,
                          (demo::currency, curr)
                          (int, amount))

int main()
{
    demo::employee bert = { {"bertie", "bertie's house"}, {demo::dollar, 101} };
    demo::employee ernie = { {"ernie", "ernie's house"}, {demo::dollar, 202} };
    demo::employee nelly = { {"nelly", "nelly's house"}, {demo::sterling, 303} };
    demo::employee ellie = { {"ellie", "ellie's house"}, {demo::sterling, 404} };

    demo::dept us_dept = { "us office" };
    us_dept.units.push_back(bert);
    us_dept.units.push_back(ernie);

    demo::dept uk_dept = { "uk office" };
    uk_dept.units.push_back(nelly);

    demo::dept uk_accounts = { "uk accounts dept" };
    uk_accounts.units.push_back(ellie);
    uk_dept.units.push_back(uk_accounts);

    demo::company c;
    c.push_back(us_dept);
    c.push_back(uk_dept);

    // Raise everybodies salaries by 1 unit. Hurray!
    traversal::full(c, traversal::extend(demo::increase_salary));
    return 0;
}
