/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#include <boost/detail/lightweight_test.hpp>
#include <boost/traversal/functional/restrict.hpp>
#include <boost/utility/result_of.hpp>
#include <boost/mpl/assert.hpp>
#include <boost/type_traits/is_same.hpp>

using namespace boost;

namespace
{
    struct poly_unary_increment
    {
        template<typename T0>
        T0 operator()(T0& t0) const
        {
            return ++t0;
        }
    };

    struct poly_binary_increment
    {
        template<typename T0, typename T1>
        void operator()(T0& t0, T1& t1) const
        {
            ++t0;
            ++t1;
        }
    };

    struct poly_ternary_increment
    {
        template<typename T0, typename T1, typename T2>
        void operator()(T0& t0, T1& t1, T2& t2) const
        {
            ++t0;
            ++t1;
            ++t2;
        }
    };

    struct poly_bin_add
    {
        template<typename Lhs, typename Rhs>
        Lhs operator()(Lhs lhs, Rhs rhs) const
        {
            return lhs + rhs;
        }
    };

    template<typename Ret, typename MonoUnaryFunc>
    void test_result_type(MonoUnaryFunc f)
    {
        BOOST_MPL_ASSERT((is_same<typename MonoUnaryFunc::result_type, Ret>));
    }
}

int main()
{
    std::string const arg0("arg0");
    std::string const arg1("arg1");

    test_result_type<int>(traversal::restrict<int(long)>(poly_unary_increment()));
    BOOST_TEST(traversal::restrict<int(int)>(poly_unary_increment())(101) == 102);
    BOOST_TEST(traversal::restrict<int(int)>(poly_unary_increment())(arg0) == 0);
    test_result_type<int>(traversal::restrict<int(long,short)>(poly_bin_add()));
    BOOST_TEST(traversal::restrict<int(int,int)>(poly_bin_add())(101, 202) == 303);
    BOOST_TEST(traversal::restrict<int(int,int)>(poly_bin_add())(arg0, arg1) == 0);

    test_result_type<void>(traversal::restrict<void(int&)>(poly_unary_increment()));
    test_result_type<void>(traversal::restrict<void(int&,int&)>(poly_binary_increment()));
    test_result_type<void>(traversal::restrict<void(int&,int&,int&)>(poly_ternary_increment()));

    int i0 = 0;
    traversal::restrict<void(int&)>(poly_unary_increment())(i0);
    BOOST_TEST(i0 == 1);

    int j0 =0, j1 = 0;
    traversal::restrict<void(int&,int&)>(poly_binary_increment())(j0,j1);
    BOOST_TEST(j0 == 1);
    BOOST_TEST(j1 == 1);

    int k0 = 0, k1 = 0, k2 = 0;
    traversal::restrict<void(int&,int&,int&)>(poly_ternary_increment())(k0,k1,k2);
    BOOST_TEST(k0 == 1);
    BOOST_TEST(k1 == 1);
    BOOST_TEST(k2 == 1);

    return boost::report_errors();
}
