/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#include <boost/detail/lightweight_test.hpp>
#include <boost/traversal/functional/tu_restrict.hpp>

#include <utility>

namespace
{
    struct triple
    {
        template<typename U0, typename U1>
        U0 operator()(U0& u0, U1& u1) const
        {
            u0.first += u1 * 3;
            return u0;
        }
    };

    struct add
    {
        template<typename U0, typename U1, typename U2>
        U0 operator()(U0& u0, U1& u1, U2& u2) const
        {
            u0.first += u1 + u2;
            return u0;
        }
    };

}

int main()
{
    namespace traversal = boost::traversal;
    int one = 1, two = 2;
    std::string dummy1 = "dummy1", dummy2 = "dummy2";
    typedef std::pair<int, bool> state_type;
    {
        state_type state(0, false);
        BOOST_TEST(
           traversal::tu_restrict<state_type(state_type&, int&)>(triple())(state, one).first == 3);
    }

    {
        state_type state(0, false);
        BOOST_TEST(
           traversal::tu_restrict<state_type(state_type&, int&)>(triple())(state, dummy1).first == 0);
    }

    {
        state_type state(0, false);
        BOOST_TEST(
            traversal::tu_restrict<state_type(state_type&, int&, int&)>(add())(state, one, two).first == 3);
    }

    {
        state_type state(0, false);
        BOOST_TEST(
            traversal::tu_restrict<state_type(state_type&, int&, int&)>(add())(state, dummy1, dummy2).first == 0);
    }
    return boost::report_errors();
}
