/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#include <boost/detail/lightweight_test.hpp>
#include <boost/traversal/functional/unstructured.hpp>
#include <boost/traversal/recursive/full.hpp>
#include <boost/traversal/horizontal/all.hpp>
#include <boost/traversal/horizontal/tu_all.hpp>
#include <boost/traversal/recursive/tu_full.hpp>

#include <vector>

using namespace boost;

namespace
{
    struct increment_all
    {
        bool operator()(int& i) const
        {
            ++i;
            return false;
        }

        template<typename U>
        bool operator()(U&) const
        {
            return false;
        }
    };

    struct tu_add_all_ints
    {
        template<typename State, typename U>
        State operator()(State state, U&) const
        {
            return state;
        }

        template<typename State>
        State operator()(State state, int& i) const
        {
            state.first += i;
            return state;
        }
    };
}

int main()
{
    typedef std::vector<int> int_vec;

    std::vector<int_vec > v0(1, int_vec(1, 101));
    traversal::full(v0, increment_all());
    BOOST_TEST(v0[0][0] == 102);

    std::vector<int_vec > v1(1, int_vec(1, 101));
    traversal::full(v1, increment_all(), traversal::unstructured<int_vec >(traversal::all));
    BOOST_TEST(v1[0][0] == 101);

    int_vec v3(2, 101);
    std::pair<int, bool> const v3_res = traversal::tu_full(v3, 1, tu_add_all_ints());
    BOOST_TEST(v3_res.first == 1 + 2 * 101);

    int_vec v4(2, 101);
    std::pair<int, bool> const v4_res = 
        traversal::tu_full(v4, 1, tu_add_all_ints(), traversal::tu_unstructured<int_vec>(traversal::tu_all));
    BOOST_TEST(v4_res.first == 1);

    return boost::report_errors();
}
