/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#include <boost/detail/lightweight_test.hpp>
#include <boost/traversal/horizontal/reverse_all.hpp>

#include <vector>
#include <memory>
#include <boost/fusion/container/vector/vector10.hpp>
#include <boost/fusion/sequence/intrinsic/at.hpp>
#include <boost/optional.hpp>

using namespace boost;

namespace
{
    struct increment_always
    {
        template<typename T>
        bool operator()(T& t) const
        {
            ++t;
            return false;
        }
    };

    struct increment_once
    {
        template<typename T>
        bool operator()(T& t) const
        {
            ++t;
            return true;
        }
    };

    void reverse_all_tests()
    {
        int i = 0;
        traversal::reverse_all(i, increment_always());
        BOOST_TEST(i == 0);

        int* i_ptr = &i;
        traversal::reverse_all(i_ptr, increment_always());
        BOOST_TEST(i == 1);

        std::auto_ptr<int> a_ptr(new int(101));
        traversal::reverse_all(a_ptr, increment_always());
        BOOST_TEST(*a_ptr == 102);

        std::vector<int> std_v(2, 101);
        traversal::reverse_all(std_v, increment_always());
        BOOST_TEST(std_v[0] == 102);
        BOOST_TEST(std_v[1] == 102);

        fusion::vector2<int,long> fusion_v(101, 202);
        traversal::reverse_all(fusion_v, increment_always());
        BOOST_TEST(fusion::at_c<0>(fusion_v) == 102);
        BOOST_TEST(fusion::at_c<1>(fusion_v) == 203);

        optional<int> o(101);
        traversal::reverse_all(o, increment_always());
        BOOST_TEST(o.get() == 102);
    }

    void once_tests()
    {
        int i = 0;
        traversal::reverse_all(i, increment_once());
        BOOST_TEST(i == 0);

        int* i_ptr = &i;
        traversal::reverse_all(i_ptr, increment_once());
        BOOST_TEST(i == 1);

        std::auto_ptr<int> a_ptr(new int(101));
        traversal::reverse_all(a_ptr, increment_once());
        BOOST_TEST(*a_ptr == 102);

        std::vector<int> std_v(2, 101);
        traversal::reverse_all(std_v, increment_once());
        BOOST_TEST(std_v[0] == 102);
        BOOST_TEST(std_v[1] == 102);

        fusion::vector2<int,long> fusion_v(101, 202);
        traversal::reverse_all(fusion_v, increment_once());
        BOOST_TEST(fusion::at_c<0>(fusion_v) == 102);
        BOOST_TEST(fusion::at_c<1>(fusion_v) == 203);

        optional<int> o(101);
        traversal::reverse_all(o, increment_once());
        BOOST_TEST(o.get() == 102);
    }
}

int main()
{
    reverse_all_tests();
    once_tests();
    return boost::report_errors();
}
