/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#include <boost/detail/lightweight_test.hpp>
#include <boost/traversal/horizontal/tu_reverse_one.hpp>

#include <vector>
#include <memory>
#include <boost/fusion/container/vector/vector10.hpp>
#include <boost/fusion/sequence/intrinsic/at.hpp>
#include <boost/optional.hpp>
#include <boost/variant.hpp>

using namespace boost;

namespace
{
    struct tu_all_add_ints
    {
        std::pair<int, bool>
        operator()(std::pair<int, bool> lhs, int rhs) const
        {
            lhs.first += rhs;
            return lhs;
        }

        template<typename T>
        std::pair<int, bool>
        operator()(std::pair<int, bool> lhs, T const&) const
        {
            return lhs;
        }
    };

    struct tu_once_add_ints
    {
        std::pair<int, bool>
        operator()(std::pair<int, bool> lhs, int rhs) const
        {
            return std::make_pair(lhs.first + rhs, true);
        }

        template<typename T>
        std::pair<int, bool>
        operator()(std::pair<int, bool> lhs, T const&) const
        {
            return lhs;
        }
    };

    void all_tests()
    {
        std::vector<int> v(2, 101);
        std::pair<int, bool> const vec_res = traversal::tu_reverse_one(v, tu_all_add_ints(), 1);
        BOOST_TEST(vec_res.first == 1 + 2* 101);
        BOOST_TEST(!vec_res.second);

        optional<int> o(101);
        std::pair<int, bool> const o_res = traversal::tu_reverse_one(o, tu_all_add_ints(), 1);
        BOOST_TEST(o_res.first == 1 + 101);
        BOOST_TEST(!o_res.second);

        std::auto_ptr<int> a(new int(101));
        std::pair<int, bool> const a_res = traversal::tu_reverse_one(a, tu_all_add_ints(), 1);
        BOOST_TEST(a_res.first == 1 + 101);
        BOOST_TEST(!a_res.second);

        variant<int, char> var(101);
        std::pair<int, bool> const var_res = traversal::tu_reverse_one(var, tu_all_add_ints(), 1);
        BOOST_TEST(var_res.first == 1 + 101);
        BOOST_TEST(!var_res.second);

        fusion::vector2<int, int> v2(101, 202);
        std::pair<int, bool> const v2_res = traversal::tu_reverse_one(v2, tu_all_add_ints(), 1);
        BOOST_TEST(v2_res.first == 1 + 101 + 202);
        BOOST_TEST(!v2_res.second);
    }

    void once_tests()
    {
        std::vector<int> v;
        v.push_back(101);
        v.push_back(202);
        std::pair<int, bool> const vec_res = traversal::tu_reverse_one(v, tu_once_add_ints(), 1);
        BOOST_TEST(vec_res.first == 1 + 202);
        BOOST_TEST(vec_res.second);

        optional<int> o(101);
        std::pair<int, bool> const o_res = traversal::tu_reverse_one(o, tu_once_add_ints(), 1);
        BOOST_TEST(o_res.first == 1 + 101);
        BOOST_TEST(o_res.second);

        std::auto_ptr<int> a(new int(101));
        std::pair<int, bool> const a_res = traversal::tu_reverse_one(a, tu_once_add_ints(), 1);
        BOOST_TEST(a_res.first == 1 + 101);
        BOOST_TEST(a_res.second);

        variant<int, char> var(101);
        std::pair<int, bool> const var_res = traversal::tu_reverse_one(var, tu_once_add_ints(), 1);
        BOOST_TEST(var_res.first == 1 + 101);
        BOOST_TEST(var_res.second);

        fusion::vector2<int, int> v2(101, 202);
        std::pair<int, bool> const v2_res = traversal::tu_reverse_one(v2, tu_once_add_ints(), 1);
        BOOST_TEST(v2_res.first == 1 + 202);
        BOOST_TEST(v2_res.second);
    }
}

int main()
{
    all_tests();
    once_tests();
    return boost::report_errors();
}
