/*=============================================================================
    Copyright (c) 2007-2008 Dan Marsden

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#include <boost/traversal/recursive/bu_full.hpp>
#include <boost/detail/lightweight_test.hpp>

#include <sstream>
#include <vector>
#include <boost/fusion/container/vector/vector10.hpp>
#include <boost/fusion/sequence/intrinsic/at.hpp>

using namespace boost;

namespace
{
    struct stream_ints
    {
        explicit stream_ints(std::ostream& stream)
            : stream(stream)
        {}

        void operator()(int& i) const
        {
            stream << i;
        }

        template<typename T>
        void operator()(T& t) const
        {}

        std::ostream& stream;
    };
}

int main()
{
    std::vector<int> std_v;
    std_v.push_back(101);
    std_v.push_back(202);
    std::ostringstream std_stream;
    traversal::bu_full(std_v, stream_ints(std_stream));
    BOOST_TEST(std_stream.str() == "101202");

    fusion::vector2<int, int> fusion_v(101, 202);
    std::ostringstream fusion_stream;
    traversal::bu_full(fusion_v, stream_ints(fusion_stream));
    BOOST_TEST(fusion_stream.str() == "101202");

    return boost::report_errors();
}
